/*
 * Debug Sub-System (DebugSS) module for Keystone devices
 *
 * Copyright (C) 2014 Texas Instruments.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include <linux/module.h>
#include <linux/sysfs.h>
#include <linux/err.h>
#include <asm/io.h>
#include <linux/sched.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/init.h>
#include <linux/platform_device.h>
#include <linux/clk.h>
#include <linux/mm.h>
#include <linux/miscdevice.h>
#include <linux/slab.h>
#include "debugss_kmodule.h"

/* un-comment the line below to enable printing for kernel debug messages */
//#define DEBUGSS_DRV_DEBUG

#ifdef DEBUGSS_DRV_DEBUG
#define __D(fmt, args...) dev_info(debugss_device, "debug_info: " fmt, ## args)
#else
#define __D(fmt, args...)
#endif

#define __E(fmt, args...) dev_err(debugss_device, "drv_error: " fmt, ## args)

/* static variable for storing the current device type */
static edevice_type target_device_type; 

/* Device driver static variables */
static struct device* debugss_device = NULL;
static struct clk		*mainpll_clk;
static struct clk		*armpll_clk;
static struct clk		*gemtrace_clk;
static struct clk		*debugss_clk;
static struct clk		*tetb_clk;

/* Address map linked list node */
struct addr_map_node_t {
        struct vm_area_struct *vma;
        pid_t process_id;
        uint32_t phy_addr_start;
        uint32_t phy_addr_end;
        struct addr_map_node_t * next_node;
        struct addr_map_node_t * prev_node;
};

/* static head pointer for address map linked list */
static struct addr_map_node_t * head_ptr = NULL;   

static edevice_type identify_device(uint32_t jtagid)
{
        __D("%s:\n",__FUNCTION__);    

        jtagid = (jtagid & 0x0FFFF000) >> 12;

        if(jtagid == TCI6614_JTAGID_PARTNUMBER) {
                return(eTCI6614_device);
        } else if(jtagid == C66AK2Hxx_JTAGID_PARTNUMBER) {
                return(e66AK2Hxx_device);
        } else if(jtagid == C66AK2Exx_JTAGID_PARTNUMBER) {
                return(e66AK2Exx_device);
        } else if(jtagid == TCI6630K2L_JTAGID_PARTNUMBER) {
                return(eTCI6630K2L_device);
        }
	else{
                return(eNotSupported_device);
        }
}

/* This fuction returns 1 if the input address range is a valid debugss 
address range or else returns -1 for unvalid debugss address range */
static int check_valid_debugss_addr_range(uint32_t range_start_addr, 
                                          uint32_t range_end_addr)
{
        uint32_t idx;

        __D("%s:\n",__FUNCTION__);

        if(e66AK2Hxx_device == target_device_type) {
                for(idx=0; idx<eK2H_DEBUGSS_NUM_RANGE_MAX; idx++) {
                        if((range_start_addr >= k2h_debugss_addr_range[idx][0]) 
                       && (range_end_addr <= k2h_debugss_addr_range[idx][1])) {
                                /* valid debugss address range */                                
                                return(1);
                        }
                }

                if(eK2H_DEBUGSS_NUM_RANGE_MAX == idx){
                        return(-1); /* unvalid debugss address range */
                }
        }

        return(-1);
}

static int range_is_allowed(unsigned long pfn, unsigned long size)
{
        return 1;
}

int __weak phys_mem_access_prot_allowed(struct file *file,
        unsigned long pfn, unsigned long size, pgprot_t *vma_prot)
{
        return 1;
}

#ifndef __HAVE_PHYS_MEM_ACCESS_PROT

/*
 * Architectures vary in how they handle caching for addresses
 * outside of main memory.
 *
 */
#ifdef pgprot_noncached
static int uncached_access(struct file *file, phys_addr_t addr)
{
        /*
         * Accessing memory above the top the kernel knows about or through a
         * file pointer
         * that was marked O_DSYNC will be done non-cached.
         */
        if (file->f_flags & O_DSYNC)
                return 1;
        return addr >= __pa(high_memory);
}
#endif

static pgprot_t phys_mem_access_prot(struct file *file, unsigned long pfn,
                                     unsigned long size, pgprot_t vma_prot)
{
#ifdef pgprot_noncached
        phys_addr_t offset = pfn << PAGE_SHIFT;

        if (uncached_access(file, offset))
                return pgprot_noncached(vma_prot);
#endif
        return vma_prot;
}
#endif

static int private_mapping_ok(struct vm_area_struct *vma)
{
        return 1;
}

static int debugss_device_open(struct inode* inode, struct file* filp)
{
	__D("%s:\n",__FUNCTION__);

	/* Do nothing */

	return 0;
}

static int debugss_device_close(struct inode* inode, struct file* filp)
{
	__D("%s:\n",__FUNCTION__);

	/* Do nothing */

	return 0;
}

void debugss_vma_open(struct vm_area_struct *vma)
{   
	__D("%s:Phy address:0x%lx, Virt address:0x%lx, size:%d\n",__FUNCTION__,
		vma->vm_pgoff << PAGE_SHIFT, vma->vm_start, 
                (int)(vma->vm_end - vma->vm_start));

	return;
}
void debugss_vma_close(struct vm_area_struct *vma)
{
	struct addr_map_node_t * current_ptr = NULL;
	unsigned int phy_addr_start, phy_addr_end;
	phy_addr_start = vma->vm_pgoff << PAGE_SHIFT;
	phy_addr_end = (vma->vm_pgoff << PAGE_SHIFT) + (vma->vm_end - vma->vm_start) - sizeof(uint32_t);    
	__D("%s:Phy address:0x%x, Virt address:0x%lx, size:%d\n",__FUNCTION__, 
	    phy_addr_start, vma->vm_start, (int)(vma->vm_end - vma->vm_start));

	if(NULL != head_ptr) {
		current_ptr = head_ptr;

		while(1){
			if((current_ptr->vma == vma) && 
			(current_ptr->phy_addr_start == phy_addr_start) &&
			(current_ptr->phy_addr_end == phy_addr_end) &&
			(current_ptr->process_id == current->pid)) {
				/* Remove the current node from the list */
				/* check if this is the only node in the list */
				if((NULL == current_ptr->prev_node) && 
				   (NULL == current_ptr->next_node)) {
					head_ptr = NULL;
				} else if(NULL == current_ptr->prev_node) {
					/* first node of the list */
					current_ptr->next_node->prev_node = NULL;
				} else if(NULL == current_ptr->next_node) {
		                	/* last node of the list */
					current_ptr->prev_node->next_node = NULL;
				} else {
					/* any node in the middle */
					current_ptr->prev_node->next_node = current_ptr->next_node;
					current_ptr->next_node->prev_node = current_ptr->prev_node;
				}

				/* Free current node */
				kfree((const void *)current_ptr);

				__D("%s: Removed node successfully\n",__FUNCTION__);

				break;
			}

			if(NULL == current_ptr->next_node) {
				break;
			} else {
				current_ptr = current_ptr->next_node;
			}
		}
	}

	return;
}

static struct vm_operations_struct debugss_remap_vm_ops = {
        .open = debugss_vma_open,
        .close = debugss_vma_close,
};

static const struct vm_operations_struct mmap_mem_ops = {
#ifdef CONFIG_HAVE_IOREMAP_PROT
        .access = generic_access_phys
#endif
};

static int debugss_remap_mmap(struct file *filp, struct vm_area_struct *vma)
{
	size_t size;
	unsigned int phy_addr_start, phy_addr_end;
	struct addr_map_node_t * current_ptr = NULL;
	struct addr_map_node_t * new_ptr = NULL;
	__D("%s:\n",__FUNCTION__);

	size = vma->vm_end - vma->vm_start;

	if (!private_mapping_ok(vma))
	    return -ENOSYS;

	if (!range_is_allowed(vma->vm_pgoff, size))
	    return -EPERM;

	if (!phys_mem_access_prot_allowed(filp, vma->vm_pgoff, size,
		                            &vma->vm_page_prot))
	    return -EINVAL;


	vma->vm_page_prot = phys_mem_access_prot(filp, vma->vm_pgoff,
		                             size,
		                             vma->vm_page_prot);

	vma->vm_ops = &mmap_mem_ops;

	/* check if the requested PHY address range to be mapped is part
	of the device debugss memory map */
	phy_addr_start = vma->vm_pgoff << PAGE_SHIFT;
	phy_addr_end = (vma->vm_pgoff << PAGE_SHIFT) + size - sizeof(uint32_t);
	__D("%s:Phy address start:0x%lx, Phy address end:0x%lx, size:%d\n",__FUNCTION__,phy_addr_start,phy_addr_end, size);
	if(-1 == check_valid_debugss_addr_range(phy_addr_start, phy_addr_end)) {
		/* return error, invalid debugss address range */
		__E("%s: Invalid debugss physical address range is being provided\n",__FUNCTION__);
		return -EINVAL;
	}

	/* check if the requested PHY address range is already mapped by
	a different process */
	if(NULL != head_ptr) {
		current_ptr = head_ptr;

		while(1) {
			if(((phy_addr_start >= current_ptr->phy_addr_start) && 
			(phy_addr_start <= current_ptr->phy_addr_end)) || 
			((phy_addr_end >= current_ptr->phy_addr_start) && 
			(phy_addr_end <= current_ptr->phy_addr_end))) {
				if(current->pid != current_ptr->process_id) {
					/* return error, the requested debugss 
   				        address range is already mapped by 
					another process*/
					__E("%s: Requested debugss physical address range is already mapped by another process with pid:%d\n",__FUNCTION__,current_ptr->process_id);
					return -EINVAL;
				}
			}

			if(NULL == current_ptr->next_node) {
				break;
			} else {
				current_ptr = current_ptr->next_node;
			}
		}
	}

	/* Remap-pfn-range will mark the range VM_IO */
	if (remap_pfn_range(vma,
		        vma->vm_start,
		        vma->vm_pgoff,
		        size,
		        vma->vm_page_prot)) {
		return -EAGAIN;
	}

	vma->vm_ops = &debugss_remap_vm_ops;
	debugss_vma_open(vma);

	/* Add a new address map node to the end of the list */
	new_ptr = (struct addr_map_node_t *)kmalloc(sizeof(struct addr_map_node_t), GFP_KERNEL);
	if(NULL == new_ptr) {
		__E("%s: kmalloc returned NULL, Kernel memory allocation error\n",__FUNCTION__);
		return -ENOMEM;
	}

	//populate the new node
	new_ptr->vma = vma;
	new_ptr->process_id = current->pid;
	new_ptr->phy_addr_start = vma->vm_pgoff << PAGE_SHIFT;
	new_ptr->phy_addr_end = (vma->vm_pgoff << PAGE_SHIFT) + 
			        (vma->vm_end - vma->vm_start) - sizeof(uint32_t);

	/* Traverse to the end of the list */
	if(NULL != head_ptr) {
		current_ptr = head_ptr;

		while(NULL != current_ptr->next_node) {
			current_ptr = current_ptr->next_node;
		}
		//Add this node to end of the existing list
		new_ptr->prev_node = current_ptr;
		current_ptr->next_node = new_ptr;
		new_ptr->next_node = NULL;        
	} else {
		//Add this node as the first node in the list
		head_ptr = new_ptr;
		new_ptr->prev_node = NULL;
		new_ptr->next_node = NULL;
	}

	__D("%s: Added node successfully\n",__FUNCTION__);

	return 0;
}

static const struct file_operations debugss_fops = {
	.open = debugss_device_open,
	.mmap = debugss_remap_mmap,
	.release = debugss_device_close,
};

static struct miscdevice debugss_miscdev = {
	.minor = MISC_DYNAMIC_MINOR,
	.name = "debugss",
	.fops = &debugss_fops,
};

static ssize_t show_device_name(struct device* dev, 
				struct device_attribute* attr, 
				char* buf)
{
	ssize_t byte_count;

	__D("%s:\n",__FUNCTION__);

	if(e66AK2Hxx_device == target_device_type) {
		byte_count = sprintf(buf,"66AK2Hxx\n");
	}
	else if(eTCI6614_device == target_device_type) {
		byte_count = sprintf(buf,"TCI6614\n");
	}
	else if(eTCI6630K2L_device == target_device_type) {
		byte_count = sprintf(buf,"TCI6630K2L\n");
	}
	else if(e66AK2Exx_device == target_device_type) {
		byte_count = sprintf(buf,"66AK2Exx\n");
	}
	else {
		byte_count = sprintf(buf,"Un-Supported Device\n");
	}

	return byte_count;
}

static ssize_t get_mainpllclk(struct device* dev, 
			      struct device_attribute* attr, 
                              char* buf)
{
	ssize_t byte_count;
	unsigned long clk_rate;

	__D("%s:\n",__FUNCTION__);

	/* Get main PLL clock rate*/
	clk_rate = clk_get_rate(mainpll_clk);

	byte_count = sprintf(buf,"%ld\n",clk_rate);

	return byte_count;
}

static ssize_t get_armpllclk(struct device* dev, 
				struct device_attribute* attr, 
				char* buf)
{
	ssize_t byte_count;
	unsigned long clk_rate;

	__D("%s:\n",__FUNCTION__);

	/* Get arm PLL clock rate*/
	clk_rate = clk_get_rate(armpll_clk);

	byte_count = sprintf(buf,"%ld\n",clk_rate);

	return byte_count;
}

static ssize_t get_debugssclk(struct device* dev, 
			      struct device_attribute* attr, 
   			      char* buf)
{
	ssize_t byte_count;
	unsigned long clk_rate;

	__D("%s:\n",__FUNCTION__);

	/* Get debugss clock rate*/
	clk_rate = clk_get_rate(debugss_clk);

	byte_count = sprintf(buf,"%ld\n",clk_rate);

	return byte_count;
}

static ssize_t get_gemtraceclk(struct device* dev, 
			       struct device_attribute* attr, 
			       char* buf)
{
	ssize_t byte_count;
	unsigned long clk_rate;

	__D("%s:\n",__FUNCTION__);

	/* Get gemtrace clock rate*/
	clk_rate = clk_get_rate(gemtrace_clk);

	byte_count = sprintf(buf,"%ld\n",clk_rate);

	return byte_count;
}

static ssize_t get_tetbclk(struct device* dev, 
		           struct device_attribute* attr, 
			   char* buf)
{
	ssize_t byte_count;
	unsigned long clk_rate;

	__D("%s:\n",__FUNCTION__);

	/* Get tetb clock rate*/
	clk_rate = clk_get_rate(tetb_clk);

	byte_count = sprintf(buf,"%ld\n",clk_rate);

	return byte_count;

}

/* Declare the sysfs entries. The macros create instances of 
dev_attr_device_name, dev_attr_mainpllclk, dev_attr_armpllclk 
dev_attr_debugssclk and dev_attr_gemtraceclk */
static DEVICE_ATTR(device_name, S_IRUGO, show_device_name, NULL);
static DEVICE_ATTR(mainpllclk, S_IRUGO, get_mainpllclk, NULL);
static DEVICE_ATTR(armpllclk, S_IRUGO, get_armpllclk, NULL);
static DEVICE_ATTR(debugssclk, S_IRUGO, get_debugssclk, NULL);
static DEVICE_ATTR(gemtraceclk, S_IRUGO, get_gemtraceclk, NULL);
static DEVICE_ATTR(tetbclk, S_IRUGO, get_tetbclk, NULL);

static int keystone_debugss_probe(struct platform_device *pdev)
{
	void * bootcfg_reg_map_ptr;
	unsigned int jtagid;
	int retval = 0;

	__D("%s:\n",__FUNCTION__);

	debugss_device = &pdev->dev;

	/* Prepare mainpll clock */
	mainpll_clk = devm_clk_get(debugss_device, "mainpllclock");
	if (WARN_ON(IS_ERR(mainpll_clk)))
		return PTR_ERR(armpll_clk);

	clk_prepare_enable(mainpll_clk);

	/* Prepare armpll clock */
	armpll_clk = devm_clk_get(debugss_device, "armpllclock");
	if (WARN_ON(IS_ERR(armpll_clk)))
		return PTR_ERR(mainpll_clk);

	clk_prepare_enable(armpll_clk);

	/* Prepare gemtrace clock */
	gemtrace_clk = devm_clk_get(debugss_device, "gemtraceclock");
	if (WARN_ON(IS_ERR(gemtrace_clk)))
		return PTR_ERR(gemtrace_clk);

	clk_prepare_enable(gemtrace_clk);

	/* Prepare debugss clock */
	debugss_clk = devm_clk_get(debugss_device, "debugssclock");
	if (WARN_ON(IS_ERR(debugss_clk)))
		return PTR_ERR(debugss_clk);

	clk_prepare_enable(debugss_clk);

	/* Prepare tetb clock */
	tetb_clk = devm_clk_get(debugss_device, "tetbclock");
	if (WARN_ON(IS_ERR(tetb_clk)))
		return PTR_ERR(tetb_clk);

	clk_prepare_enable(tetb_clk);

	/* Create debugss misc device */
	retval = misc_register(&debugss_miscdev);
	if (retval < 0) {
		dev_err(debugss_device, "cannot register debugss misc device\n");
	}

	debugss_device = debugss_miscdev.this_device;

	/* Identify target device type by reading the JTAG ID */
	/* Map first 4K of BOOTCFG register memory to kernel space memory */
	bootcfg_reg_map_ptr = ioremap_nocache(BOOTCFG_BASE_ADDR, PAGE_SIZE);

	/* Read JTAG ID */
	jtagid = ioread32((void *)(((char *)bootcfg_reg_map_ptr) + 
					BOOTCFG_JTAGID_REG));

	/* identify device type */
	target_device_type = identify_device(jtagid);

	/* Un-map BOOTCFG register memory from kernel space memory */
	iounmap(bootcfg_reg_map_ptr);

	/* Now we can create the sysfs endpoints */
	retval = device_create_file(debugss_device, &dev_attr_device_name);
	if (retval < 0) {
		__E("failed to create debugss device_name /sys endpoint\n");
	}
	retval = device_create_file(debugss_device, &dev_attr_mainpllclk);
	if (retval < 0) {
		__E("failed to create debugss mainpllclk /sys endpoint\n");
	}
	retval = device_create_file(debugss_device, &dev_attr_armpllclk);
	if (retval < 0) {
		__E("failed to create debugss armpllclk /sys endpoint\n");
	}
	retval = device_create_file(debugss_device, &dev_attr_debugssclk);
	if (retval < 0) {
	__E("failed to create debugss debugssclk /sys endpoint\n");
	}
	retval = device_create_file(debugss_device, &dev_attr_gemtraceclk);
	if (retval < 0) {
		__E("failed to create debugss gemtraceclk /sys endpoint\n");
	}
	retval = device_create_file(debugss_device, &dev_attr_tetbclk);
	if (retval < 0) {
		__E("failed to create debugss tetbclk /sys endpoint\n");
	}

	return retval;
}

static int keystone_debugss_remove(struct platform_device *pdev)
{
	__D("%s:\n",__FUNCTION__);

	device_remove_file(debugss_device, &dev_attr_device_name);
	device_remove_file(debugss_device, &dev_attr_mainpllclk);
	device_remove_file(debugss_device, &dev_attr_armpllclk);
	device_remove_file(debugss_device, &dev_attr_debugssclk);
	device_remove_file(debugss_device, &dev_attr_gemtraceclk);
	device_remove_file(debugss_device, &dev_attr_tetbclk);

	/* De-register debugss misc device */
	misc_deregister(&debugss_miscdev);

	/* Un-prepare mainpll clock */
	clk_disable_unprepare(mainpll_clk);

	/* Un-prepare armpll clock */
	clk_disable_unprepare(armpll_clk);

	/* Un-prepare gemtrace clock */
	clk_disable_unprepare(gemtrace_clk);

	/* Un-prepare debugSS clock */
	clk_disable_unprepare(debugss_clk);

	/* Un-prepare TETB clock */
	clk_disable_unprepare(tetb_clk);

	return 0;
}

static const struct of_device_id keystone_debugss_of_match[] = {
	{ .compatible = "ti,keystone-debugss", },
	{},
};
MODULE_DEVICE_TABLE(of, keystone_debugss_of_match);

static struct platform_driver platform_debugss_driver = {
	.driver = {
		.name = "debugss",
		.owner	= THIS_MODULE,
		.of_match_table = keystone_debugss_of_match,
	},
	.probe = keystone_debugss_probe,
	.remove = keystone_debugss_remove,
};

module_platform_driver(platform_debugss_driver);

/* Module information */
MODULE_AUTHOR("Texas Instruments Incorporated");
MODULE_DESCRIPTION("Debug Sub-System (DebugSS) module for Keystone devices");
MODULE_LICENSE("GPL v2");
MODULE_ALIAS("platform:" DRIVER_NAME);
